@@ -17,11 +17,17 @@
 ;; GNU General Public License for more details.
 ;;
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
+;; along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Code:
 (require 'cl-lib)
 
+(defcustom wisi-incremental-parse-enable nil
+  "If non-nil, use incremental parse when possible."
+  :type 'boolean
+  :group 'wisi
+  :safe #'booleanp)
+
 (defcustom wisi-partial-parse-threshold 100001
   "Minimum size that will be parsed by each call to the parser.
 A parse is always requested at a point (or on a region); the
@@ -33,7 +39,7 @@ is properly indented. Most navigate parses ignore this setting
 and parse the whole buffer."
   :type 'integer
   :group 'wisi
-  :safe 'integerp)
+  :safe #'integerp)
 (make-variable-buffer-local 'wisi-partial-parse-threshold)
 
 (defcustom wisi-save-all-changes nil
@@ -46,6 +52,16 @@ for the changes. The filename is the visited file name with
   :group 'wisi
   :safe 'booleanp)
 
+(defcustom wisi-save-text-tree nil
+  "When non-nil, save the parser's copy of the full text and the
+current parse tree before each change, to aid in reproducing
+bugs. The full text is written to a file whose name is the
+visited file name with \"-wisi-prev-text\"' appended. The tree
+can be dumped to a file via the tree query dump-prev."
+  :type 'boolean
+  :group 'wisi
+  :safe #'booleanp)
+
 (cl-defstruct (wisi--lexer-error)
   pos ;; position (integer) in buffer where error was detected.
   message  ;; string error message
@@ -146,6 +162,9 @@ region BEGIN END that starts and ends at points the parser can
 handle gracefully."
   (cons begin end))
 
+(cl-defgeneric wisi-parse-save-text-tree-auto (parser enable)
+  "Implement `wisi-save-text-tree'.")
+
 (defvar-local wisi-parser-shared nil
   "The current shared wisi parser; a ‘wisi-parser-shared’ object.
 There is one parser object per language; `wisi-parser-shared' is a
@@ -154,6 +173,13 @@ buffer-local reference to that shared object.")
 (defvar-local wisi-parser-local nil
   "Buffer-local values used by the wisi parser; a ‘wisi-parser-local’ object.")
 
+(defvar-local wisi--changes nil
+  "Cached list of args to wisi-after-change, for incremental parse.
+Each element is
+(INSERT-BEGIN-BYTE-POS INSERT-BEGIN-CHAR-POS
+ INSERT-END-BYTE-POS INSERT-END-CHAR-POS
+ DELETED-BYTE-COUNT DELETED-CHAR-COUNT INSERTED-TEXT)")
+
 (defconst wisi-post-parse-actions '(navigate face indent none refactor query debug)
   "Actions that the parser can perform after parsing.
 Only navigate thru indent are valid for partial parse.")
@@ -240,7 +266,7 @@ have been previously parsed by `wisi-parse-current' or
     (parent		    .	3)
     (child		    .	4)
     (print		    .	5)
-    (dump		    .	6)
+    (dump		    .	6) ;; dump-prev is not here because it queries a different tree.
     )
   "Query values for `wisi-parse-tree-query'.")
 
@@ -436,7 +462,7 @@ Assumes the buffer is fully parsed."
 1 : report parse errors (for running tests)"
   :type 'integer
   :group 'wisi
-  :safe 'integerp)
+  :safe #'integerp)
 
 ;; The following parameters are easily changeable for debugging.
 (defcustom wisi-parser-verbosity ""
@@ -446,7 +472,7 @@ Examples:
 debug=1 lexer=1 parse=2 action=3"
   :type 'string
   :group 'wisi
-  :safe 'stringp)
+  :safe #'stringp)
 (make-variable-buffer-local 'wisi-parser-verbosity)
 
 (defcustom wisi-mckenzie-zombie-limit nil
@@ -457,7 +483,7 @@ value gives better solutions, but may cause too many parsers to
 be active at once.  If nil, uses %mckenzie_zombie_limit value from grammar file."
   :type 'integer
   :group 'wisi
-  :safe 'integerp)
+  :safe #'integerp)
 (make-variable-buffer-local 'wisi-mckenzie-zombie-limit)
 
 (defcustom wisi-mckenzie-enqueue-limit nil
@@ -467,7 +493,7 @@ Higher value has more recover power, but will be slower to fail.
 If nil, uses %mckenzie_enqueue_limit value from grammar file."
   :type 'integer
   :group 'wisi
-  :safe 'integerp)
+  :safe #'integerp)
 (make-variable-buffer-local 'wisi-mckenzie-enqueue-limit)
 
 (defcustom wisi-parse-max-parallel nil
@@ -478,7 +504,7 @@ the grammar has excessive conflicts. If nil, uses %max_parallel
 value from grammar file (default 15)"
   :type 'integer
   :group 'wisi
-  :safe 'integerp)
+  :safe #'integerp)
 (make-variable-buffer-local 'wisi-parse-max-parallel)
 
 ;; end of easily changeable parameters
@@ -504,7 +530,7 @@ Normally set from a language-specific option.")
      "%s:%d:%d: %s"
        (buffer-name) ;; buffer-file-name is sometimes nil here!?
        line col
-       (apply 'format message args))))
+       (apply #'format message args))))
 
 (defvar wisi-parse-error nil)
 (put 'wisi-parse-error
@@ -591,23 +617,6 @@ Signals an error if `wisi-incremental-parse-enable' is nil."
     (user-error "wisi-parse-incremental-none with wisi-incremental-parse-enable nil"))
   (save-excursion (wisi-parse-incremental wisi-parser-shared 'none)))
 
-(defun wisi-replay-kbd-macro (macro)
-  "Replay keyboard macro MACRO into current buffer,
-with incremental parse after each key event."
-  (unless wisi-incremental-parse-enable
-    (user-error "wisi-incremental-parse-enable nil; use EMACS_SKIP_UNLESS"))
-  (let ((i 0))
-    (while (< i  (length macro))
-      (execute-kbd-macro (make-vector 1 (aref macro i)))
-      (save-excursion
-	(condition-case err
-	    (wisi-parse-incremental wisi-parser-shared 'none)
-	  (wisi-parse-error
-	   (when (< 0 wisi-debug)
-	     ;; allow continuing when parser throws parse-error
-	     (signal (car err) (cdr err))))))
-      (setq i (1+ i)))))
-
 (defun wisi-replay-undo (count)
   "Execute `undo' COUNT times, delaying in between each."
   (let ((i 0))
